/*
			Copyright (c) 1993 by
			Advanced Visual Systems Inc.
			All Rights Reserved
	
	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.
	
	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/con_tri.c#1 $
*/
#include <avs/util.h>
#include <avs/err.h>
#include <avs/db.h>
#include <avs/fld.h>
#include <avs/math.h>
#include <avs/fld_info.h>

#define NEW_FLAG 1
#define OLD_FLAG 0

#define ERR_RETURN(A) ERRerror("cut_table",1,ERR_ORIG, A); return(0);
#define MAX_NAME_SIZE 1024

#define MAX_NSETS 6

static char *cell_name[]= {"Tet", "Prism", "Pyr", "Hex", "Tri", "Quad"};	

static FILE   *f_tbl, *f_val, *f_on, *f_nn, *f_nt, *f_min, *f_max, *f_fl, *f_con, *f_mind, *f_cind, *f_tind, *f_ct, *f_cn;
static int   m_ind, c_ind, t_ind, first=1, count = 0;
static int   on1=1, nn1=1, nt1=1, min1=1, fl1=1, con1=1, mind1=1, cind1=1, ct1=1;
static int   on0=0, nn0=0, nt0=0, min0=0, fl0=0, con0=0, mind0=0, cind0=0, ct0=0;

static int  ndn[4] = {1,2,4,8};
static int  table[8][8];
static int get_dist();

con_table(elem_id)
    OMobj_id elem_id;
{
	OMobj_id in, out, level_id, param;
	double  level;
	
	in = OMfind_subobj(elem_id, OMstr_to_name("in"), OM_OBJ_RD);
	if (FUNCcon_table(in, out)) {
		return(1);
	}
	else return(0);
}

int FUNCcon_table (in, out)
    OMobj_id  in, out;
{
	int   nnodes, nspace, out_nnodes, data_type, size, veclen, ncomp;
	int   *node_list, min_node_list[128], max_node_list[128], stat;
	int   comp, *node_conn, ppnts, pconn[10], tlist[20], flag[20];
	int   ntet, *nlist, table_index, table_index2, n, n0, n1,nt, i, j, nn, max_tet;
	int   nfaces, *face_nnodes, *face_node_list;
	int   tnodes, ctype[20], cnnode[20], cell;
	double null_value, val1, val2;
	char  units[MAX_NAME_SIZE];
	float  *coord, *out_coord;
	float   *node_data, xyz[1000], point[3], data[100], dist;
	float  min_ext[3], max_ext[3];
	char   *cell_table, *block_table;
	OMobj_id   cell_set;
	Field_Table    *ct;

	int  old_nnodes, new_nnodes, new_nnodes_r, *min_node_list_r, *max_node_list_r, *old_nlist;
	int  out_ncells[MAX_NSETS], *out_nlist[MAX_NSETS], conn_size[MAX_NSETS];
	int  tbl_ind, ind1, ind2, found, val_flag[128];
	unsigned char *list;

	null_value = 0.0;
	out_nnodes = 0;

	tbl_ind = 1;
	m_ind=1;
	c_ind=1;
	t_ind=1;

	if (!(f_tbl = fopen("f_tbl.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_val = fopen("f_val.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_on = fopen("f_on.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_nn = fopen("f_nn.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_nt = fopen("f_nt.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_min = fopen("f_min.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_max = fopen("f_max.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_fl = fopen("f_fl.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_con = fopen("f_con.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_cind = fopen("f_cind.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_mind = fopen("f_mind.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_tind = fopen("f_tind.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_ct = fopen("f_ct.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}
	if (!(f_cn = fopen("f_cn.cut", "w"))) {
		ERR_RETURN("Unable to open file");
	}

	if (FLDget_node_data_ncomp(in, &ncomp) != 1) {
		ERR_RETURN("Error geting ncomp");
	}
	if (ncomp == 0) {
		ERR_RETURN("bad number cut_table component index");
	}
	if (FLDget_node_data_veclen(in, 0, &veclen) !=1) {
		ERR_RETURN("Error geting veclen");
	}
	
	if (veclen != 1) {
		fprintf (stderr, "cut_table: first component is not scalar\n");
		return(0);
	}
	
	if (FLDget_nnodes(in, &nnodes) != 1) {
		ERR_RETURN("cannot get nnodes");
	}
	if (FLDget_nspace(in, &nspace) != 1) {
		ERR_RETURN("cannot get nnodes");
	}
	if (FLDget_node_data(in, 0, &data_type, (char **)&node_data, 
			      &size, OM_GET_ARRAY_RW) != 1) {
		ERR_RETURN("cannot get node data");
	}
	if (FLDget_coord(in, &coord, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("cannot get coordinates");
	}
	if (FLDget_coord_extent(in, min_ext, max_ext) != 1) {
		ERR_RETURN("cannot get extents");
	}
	stat = FLDget_mesh_info(in, &cell_table);
	if (!stat) {
		ERR_RETURN("cannot create cell table");
	}
	stat = FLDinterp_init(in, cell_table, &block_table);
	if (!stat) {
		ERR_RETURN("cannot create block table");
	}
	if (FLDget_cell_set(in, 0, &cell_set) != 1) {
		ERR_RETURN("cannot get cell set");
	}
	if (FLDget_node_connect(cell_set, &node_conn, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("cannot get connect");
	}

	ct = (Field_Table *)cell_table;

	fprintf (f_on, "static int CONTnold_nodes_tri[] = \n       {0,");
	fprintf (f_nn, "static int CONTnnew_nodes_tri[] = \n       {0,");
	fprintf (f_nt, "static int CONTncell_tri[] = \n       {0,");
	fprintf (f_min, "static unsigned char CONTmin_node_tri[] = \n       {0,");
	fprintf (f_max, "static unsigned char CONTmax_node_tri[] = \n       {0,");
	fprintf (f_val, "static unsigned char CONTvalues_tri[] = \n       {0,");
	fprintf (f_mind, "static int CONTm_ind_tri[] = \n       {0,");
	fprintf (f_cind, "static int CONTc_ind_tri[] = \n       {0,");
	fprintf (f_tind, "static int CONTt_ind_tri[] = \n       {0,");
	fprintf (f_con, "static unsigned char CONTconnect_list_tri[] = \n       {0,");
	fprintf (f_fl, "static unsigned char CONTnew_flag_list_tri[] = \n       {0,");
	fprintf (f_ct, "static unsigned char CONTcell_type_tri[] = \n       {0,");
	fprintf (f_cn, "static unsigned char CONTcell_nnode_tri[] = \n       {0,");

	for (n=0; n<8; n++)
		for (n0=0; n0<8; n0++)
			table[n][n0] = -1;

	max_tet=0;
	val1 = 0.5;
	val2 = 1.5;
	for (table_index2 = 1; table_index2 < 7; table_index2++) {

		for (table_index=table_index2; table_index<7; table_index++) {

			if ((table_index & table_index2) != table_index2)
				continue;
			for (n=0; n<nnodes; n++) {
				if (table_index2 & ndn[n])
					node_data[n]= 2.0;
				else if (table_index & ndn[n])
					node_data[n] = 1.0;
				else
					node_data[n] = 0.0;
			}
			UTILcontour(cell_table, node_data, data_type, 0, (char *)&null_value,
				    val1, val2,
				    &ppnts, &old_nlist,
				    &new_nnodes_r, &min_node_list_r, &max_node_list_r, &list,
				    out_ncells, out_nlist, conn_size, nspace, &out_coord);

			for (n1=0,n=0; n<nnodes; n++) {
				if (old_nlist[n] != -1) {
					pconn[n1]=n;
					memcpy(xyz+nspace*n1,out_coord+n1*nspace,nspace*sizeof(float));
					n1++;
				}
			}

			out_nnodes = 0;
			for (n=0; n<new_nnodes_r; n++) {
				/* check if inside */
				memcpy(point, out_coord+(ppnts+n)*nspace,nspace*sizeof(float));
				found = 0;
				for (n1=0; n1<nnodes; n1++) {
					if (n1 == nnodes-1) {
						ind1 = 0;
						ind2 = n1;
					}
					else {
						ind1 = n1;
						ind2 = n1+1;
					}
					get_dist(coord, nspace, ind1, ind2, point, &dist);
					if (fabs(dist) < 0.05) {
						found = 1;
						break;
					}
				}
				if (found) {
					min_node_list[out_nnodes] = ind1;
					max_node_list[out_nnodes] = ind2;
					val_flag[out_nnodes] = list[n];
					memcpy(xyz+nspace*(ppnts+out_nnodes),out_coord+(ppnts+n)*nspace,nspace*sizeof(float)); 
					out_nnodes++;
				}
			}				
			UTILdelaun_tess_2D (ppnts+out_nnodes, xyz, nspace, min_ext, max_ext, 0, 
					    &ntet, &nlist);
			if (ntet == 0)
				continue;
			table[table_index2][table_index] = tbl_ind++;

			if (on1) {
				fprintf (f_on,"%d", ppnts);
				fprintf (f_nn,"%d", out_nnodes);
				fprintf (f_mind, "%d", m_ind);
				fprintf (f_cind, "%d", c_ind);
				fprintf (f_tind, "%d", t_ind);
				on1 = 0;
			}
			else {
				fprintf (f_on,",%d", ppnts);
				fprintf (f_nn,",%d", out_nnodes);
				fprintf (f_mind, ",%d", m_ind);
				fprintf (f_cind, ",%d", c_ind);
				fprintf (f_tind, ",%d", t_ind);
			}
			on0++;
			if (on0 >= 32) {
				fprintf (f_on, ",\n        ");
				fprintf (f_nn, ",\n        ");
				fprintf (f_mind, ",\n        ");
				fprintf (f_cind, ",\n        ");
				fprintf (f_tind, ",\n        ");
				on0 = 0;
				on1 = 1;
			}
			
			for (i=0; i<out_nnodes; i++) {
			     if (min1) {
					fprintf (f_min, "%d",min_node_list[i]); 
					fprintf (f_max, "%d",max_node_list[i]); 
					fprintf (f_val, "%d",val_flag[i]); 
					m_ind++;
					min1=0;
				}
				else {
					fprintf (f_min, ",%d",min_node_list[i]); 
					fprintf (f_max, ",%d",max_node_list[i]); 
					fprintf (f_val, ",%d",val_flag[i]); 
					m_ind++;
				}
				min0++;
				if (min0 >= 24) {
					fprintf (f_min, ",\n        ");
					fprintf (f_max, ",\n        ");
					fprintf (f_val, ",\n        ");
					min0 = 0;
					min1 = 1;
				}
				
			}
			for (nt = 0,nn=0; nn<ntet; nn++) {
				for (i=0; i<3; i++) {
					if (nlist[3*nn+i]>=ppnts) {
						flag[i]=NEW_FLAG;
						tlist[i]=nlist[3*nn+i]-ppnts;
					}
					else {
						flag[i]=OLD_FLAG;
						tlist[i]=pconn[nlist[3*nn+i]];
					}
				}
				for (i=0; i<3; i++) {
					if (first) {
						fprintf (f_fl, "%d",flag[i]); 
						fprintf (f_con, "%d",tlist[i]);
						first = 0;
					}
					else {
						fprintf (f_fl, ",%d",flag[i]); 
						fprintf (f_con, ",%d",tlist[i]);
					}
					count++;
				}
				if (count >= 24) {
					fprintf (f_fl, ",\n        ");
					fprintf (f_con, ",\n        ");
					count = 0;
					first = 1;
				}
				c_ind += 3;
				nt++;
				if (ct1) {
					fprintf (f_ct, "      %d", 4);
					fprintf (f_cn, "      %d", 3);
					ct1 = 0;
				}
				else {
					fprintf (f_ct, ",%d", 4);
					fprintf (f_cn, ",%d", 3);
				}
				ct0++;
				if (ct0 >= 32) {
					fprintf (f_ct, ",\n        ");
					fprintf (f_cn, ",\n        ");
					ct0 = 0;
					ct1 = 1;
				}
				t_ind ++;
			}					
			if (nt1) {
				fprintf (f_nt, "%d", nt);
				nt1 = 0;
			}
			else {
				fprintf (f_nt, ",%d", nt);
			}
			nt0++;
			if (nt0==24) {
				fprintf (f_nt, ",\n        ");
				nt1=1;
				nt0=0;
			}
			if (nt>max_tet)
				max_tet = nt;
			/**
			  fprintf(stderr," index=%d, index1=%d, old_points=%d, new_points=%d, ntet=%d\n",
			  table_index, table_index1, ppnts, out_nnodes, nt);
			  **/
		}
	}

	fprintf (f_tbl, "static int CONTtable_index_tri[8][8] = \n       {");
	
	for (n=0; n<8; n++) {
		n1 = 1;
		for (n0=0; n0<8; n0++) {
			if (n1) {
				fprintf (f_tbl, "%d", table[n][n0]);
				n1 = 0;
			}
			else {
				fprintf (f_tbl, ",%d", table[n][n0]);
			}
		}
		fprintf (f_tbl, ",\n        ");
	}
			
	fprintf (f_tbl, "};\n\n");

	fprintf (f_on, "};\n\n");
	fprintf (f_nt, "};\n\n");
	fprintf (f_nn, "};\n\n");
	fprintf (f_min, "};\n\n");
	fprintf (f_max, "};\n\n");
	fprintf (f_val, "};\n\n");
	fprintf (f_mind, "};\n\n");
	fprintf (f_cind, "};\n\n");
	fprintf (f_tind, "};\n\n");
	fprintf (f_con, "};\n\n");
	fprintf (f_fl, "};\n\n");
	fprintf (f_ct, "};\n\n");
	fprintf (f_cn, "};\n\n");

	fclose(f_tbl);
	fclose(f_on);
	fclose(f_nn);
	fclose(f_fl);
	fclose(f_nt);
	fclose(f_min);
	fclose(f_max);
	fclose(f_val);
	fclose(f_con);
	fclose(f_mind);
	fclose(f_cind);
	fclose(f_tind);
	fclose(f_ct);
	fclose(f_cn);
	system("cat f_tbl.cut f_on.cut f_nn.cut f_mind.cut f_min.cut f_max.cut f_val.cut f_tind.cut f_ct.cut f_cn.cut f_nt.cut f_cind.cut f_con.cut f_fl.cut >cut_tbl.h");
	system("rm -rf f_*.cut");
	return(1);
}

static int print_con(nnodes,tlist, flag, nt, ctype, cell_nnode)
    int nnodes, *tlist, *flag, nt, *ctype, *cell_nnode;
{
	int  i,j;
	
	for (i=0; i<nnodes; i++) {
		if (first) {
			fprintf (f_fl, "%d",flag[i]); 
			fprintf (f_con, "%d",tlist[i]);
			first = 0;
		}
		else {
			fprintf (f_fl, ",%d",flag[i]); 
			fprintf (f_con, ",%d",tlist[i]);
		}
		count++;
	}
	if (count >= 24) {
		fprintf (f_fl, ",\n        ");
		fprintf (f_con, ",\n        ");
		count = 0;
		first = 1;
	}
	for (i=0; i<nt; i++) {
		if (ct1) {
			fprintf (f_ct, "      %d", ctype[i]);
			fprintf (f_cn, "      %d", cell_nnode[i]);
			ct1 = 0;
		}
		else {
			fprintf (f_ct, ",%d", ctype[i]);
			fprintf (f_cn, ",%d", cell_nnode[i]);
		}
		ct0 += nt;
		if (ct0 >= 32) {
			fprintf (f_ct, ",\n        ");
			fprintf (f_cn, ",\n        ");
			ct0 = 0;
			ct1 = 1;
		}
	}
	return(1);
}


static int get_dist(coord, nspace, ind1, ind2, point, dist)
    int  nspace, ind1, ind2;
    float *coord, *point, *dist;
{
	float a,b,c, *p1, *p2;
	
	p1 = coord+ind1*nspace;
	p2 = coord+ind2*nspace;
	a = p2[1]-p1[1];
	b = p1[0]-p2[0];
	c = p1[1]*(p2[0]-p1[0]) - p1[0]*(p2[1]-p1[1]);

	*dist = (a*point[0]+b*point[1]+c)/sqrt((double)(a*a+b*b));
	return(1);
}
	
